.include "tn12def.inc"

;*******************************************************************************
;*
;* A small but fine FPGA configurator
;* using a Tiny 12 and serial EEPROMs 
;*
;* Up to eigth EEPROMs can be cascaded on the I2C bus
;* All devices must be of same size
;*
;* theoritical limit is 8x64kB = 512 kB
;*
;* Examles for FPGA config size
;* Spartan-II : XC2S200(biggest family member): 166,980 Bytes
;* Spartan-3  : XC3S1000                      : 402,936 Bytes
;*
;* Configuration speed is ~8.5kBytes/s @ 2MHz
;* Top speed is ~ 43kB/s @ 10 MHz, 400kHz limit of I2C
;*
;* example uses a Xilinx XC2S30 with 42.096 bytes config size
;* and a ST24C512 with 64 kB

.equ fpga_size   = 42096;			; FPGA config file size
.equ EEPROM_size = 65536;			; EEPROM size per device
.equ F_CPU_MHz   = 2;				; CPU clock in MHz, 1..10

.equ bitreverse  = 1               

;*  = 0 , bits will be copied bit by bit from EEPROM to FPGA, faster configuration
;*        binary files must have same bitorder as required by FPGA (MSB first)
;*        For Xilinx, the bit files must be processed by bitdreher.exe!!!

;*  = 1 , bits will be copied byte by bytes from EEPROM and written bit reversed to FPGA, slow configuration
;*        binary files must have reverse bitorder as required by FPGA (LSB first)
;*        For Xilinx, the bit files are generated this way by default

;*******************************************************************************
 
.equ cfg_port = portb
.equ cfg_ddr  = ddrb
.equ cfg_pin  = pinb  

.equ DIN   = PB0      	; FPGA serial configuration
.equ DONE  = PB1
.equ CCLK  = PB2
.equ SDA   = PB3      	; Connection to I2C EEPROM
.equ SCL   = PB4

; Reset is connected to FPGA AND AVR, so we have a automatic Power-on reset

; some marcos to support readability

.macro SDA_LOW
	sbi		cfg_ddr, sda	; open drain, low
.endmacro 

.macro SDA_HIGH
	cbi		cfg_ddr, sda	; open drain, high
.endmacro

.macro SCL_LOW
	sbi		cfg_ddr, scl	; open drain, low
.endmacro 

.macro SCL_HIGH
	cbi		cfg_ddr, scl	; open drain, high
.endmacro

.macro DIN_LOW
	cbi		cfg_port, din	; push pull
.endmacro 

.macro DIN_HIGH
	sbi		cfg_port, din	; push pull
.endmacro

.macro CCLK_LOW
	cbi		cfg_port, cclk	; push pull
.endmacro 

.macro CCLK_HIGH
	sbi		cfg_port, cclk	; push pull
.endmacro

.macro i2c_error_check
	brts	PC+2
	rjmp	error_i2c
;	rcall	put_c_1
.endmacro

; add a delay to reach given delay time
; usage
; delay_1 time [Unit 0.1us], clocks_already_done
.macro delay

; calculate number of remaining clocks
; +1 is safety margin to account for integer truncation

.set clks = @0*F_CPU_MHz/10-@1+1

; use optimized delay blocks with "nonsense" commands to have maximum
; delay with minimum code size

.if clks==1
	nop
.endif

.if clks==2
	ld	null, z
.endif

.if clks==3
	lpm
.endif

.if clks==4
	lpm
	nop
.endif

.if clks==5
	lpm
	ld	null, z
.endif

.if clks==6
	lpm
	lpm
.endif

.if clks==7
	lpm
	lpm
	nop
.endif

.if clks==8
	lpm
	lpm
	ld	null,z
.endif

.if clks==9
	lpm
	lpm
	lpm
.endif

.if clks==10
	lpm
	lpm
	lpm
	nop
.endif

.if clks==11
	lpm
	lpm
	lpm
	ld	null, z
.endif

.if clks==12
	lpm
	lpm
	lpm
	lpm
.endif

.if clks==13
	lpm
	lpm
	lpm
	lpm
	nop
.endif

.if clks==14
	lpm
	lpm
	lpm
	lpm
	ld	null, z
.endif

.endmacro

.def zero  = r1			; always zero
.def one   = r2			; always one
.def null  = r3         ; write only ;-)
.def tmp1  = r16		; general purpose
.def tmp2  = r17
.def tmp3  = r18
.def cnt0  = r19        ; loop counter, 24 Bit
.def cnt1  = r20
.def cnt2  = r21			
.def i2c   = r22		; I2C address, used for multiple devices
.def abs0  = r23        ; data counter, 24 Bit
.def abs1  = r24		; 
.def abs2  = r25

;*
;*******************************************************************************

.cseg
; lets go

; initialize stack

; hardware stack for Tiny12, no init necessary

; full speed, ~2 MHz with ATtiny12

	ldi		tmp1,0xFF
	out		osccal, tmp1

; config IOs

	ldi		tmp1, (1<<cclk) | (1<<din)
	out		cfg_ddr,tmp1

	ldi		tmp1, (1<<done)
	out		cfg_port,tmp1			; internal pull up for DONE

; generate constant registers

	sub		zero,zero
	ldi		tmp1,1
	mov		one,tmp1
	ldi		i2c,0xA0				; I2C base address

; set byte counter

	ldi		abs0,byte1(fpga_size)
	ldi		abs1,byte2(fpga_size)
	ldi		abs2,byte3(fpga_size)
		
; wait 10ms for FPGA to finish power up and clear config space

	ldi		tmp1,10
	rcall	wait

; when using multiple EEPROMs, we must address them individual

device_loop:

; number of remaining bytes >EEPROM_size ?

	ldi		tmp1,byte1(EEPROM_size)
	ldi		tmp2,byte2(EEPROM_size)
	ldi		tmp3,byte3(EEPROM_size)

	cp		abs0,tmp1
	cpc		abs1,tmp2
	cpc		abs2,tmp3
	brge	device_loop_big		; >= EEPROM_size

	; less than EEPROM_size, copy remaining bytes
	mov		cnt0,abs0		; remaining bytes
	mov		cnt1,abs1
	mov		cnt2,abs2
	mov		abs0,zero		; set zero
	mov		abs1,zero
	mov		abs2,zero
	rjmp	device_loop_init

device_loop_big:
	mov		cnt0,tmp1		; EEPROM_size loop
	mov		cnt1,tmp2
	mov		cnt2,tmp3
	sub		abs0,tmp1		; - EEPROM_size
	sbc		abs1,tmp2
	sbc		abs2,tmp3

device_loop_init:
	subi	cnt0,1			; -1, since last byte must be read with no ACK
	sbci	cnt1,0			; outside the loop
	sbci	cnt2,0

; write first adress

	rcall	i2c_start
	mov		tmp1, i2c		; load i2c address
	rcall	i2c_write
	i2c_error_check
	ldi		tmp1, 0			; write adress 0
	rcall	i2c_write
	i2c_error_check
	ldi		tmp1, 0
	rcall	i2c_write
	i2c_error_check
	rcall	i2c_stop

	rcall	i2c_start
	mov		tmp1, i2c		; load i2c address
	subi	tmp1,-1			; +1, read access
	rcall	i2c_write
	i2c_error_check
	set						; set T-Flag, continous reading

main_loop:

.if bitreverse==1			
;   configuration uses online bit reverse, slow

	rcall	i2c_read
	rcall	fpga_write

.else
;   configuration uses offline bit reverse, fast (bitdreher.exe)

; read a byte from I2C and parallel send it to the FPGA
; fast version
	ldi		tmp2, 8			; loop counter

main_bit_loop:
	scl_high				; SCL rising edge
	sbic	cfg_pin,sda		; copy SDA to Din
	din_high
	sbis	cfg_pin,sda
	din_low
	delay 7, 6				; minimum LOW time 0,6us + 0.1us safty margin
	scl_low					; falling edge generates new bit
	cclk_high				; LOW time min. 9 CLKs
	cclk_low
	delay 14, 9				; minimum HIGH time 1,3us + 0.1us safty margin
	dec		tmp2
	brne	main_bit_loop

; write ACK to EEPROM

	sda_low					; write ACK, more to read
	delay  1, 0
	scl_high
	delay  7, 2
	scl_low
	sda_high
	delay  14, 10

.endif

	subi	cnt0,1			; decrement data counter
	sbci	cnt1,0
	sbci	cnt2,0
	brne	main_loop

	clt						; no ACK for last byte of block
	rcall	i2c_read
	rcall	fpga_write

	rcall	i2c_stop
	subi	i2c,-2			; +2, increment I2C address bits

; number of remaining data bytes = 0 ?

	cp		abs0,zero
	cpc		abs1,zero
	cpc		abs2,zero
	breq	main_end		; nothing left to copy, finish

	rjmp	device_loop		; process next data block

main_end:
	ldi		tmp1,0
	rcall	fpga_write		; one more dummy byte for FPGA startup

	ldi		tmp1,1
	rcall	wait			; wait for DONE to go HIGH (open drain)
	sbis	cfg_pin, done	; check if DONE is HIGH (= FPGA config OK)
	rjmp	error_fpga		; something went wrong during configuration
	ldi		tmp1, (1<<SE) | (1<<SM)		
	out		mcucr,tmp1		; power down sleep mode
	sleep					; good night
	
;  no wake up here, only through reset

; i2c has no ACK, fast blinking

error_i2c:
	sda_low
	ldi		tmp1,100
	rcall	wait
	sda_high
	ldi		tmp1,100
	rcall	wait
	rjmp	error_i2c

; FPGA configuration error, DONE stays low, slow blinking

error_fpga:
	sda_low
	ldi		tmp1,250
	rcall	wait
	ldi		tmp1,250
	rcall	wait
	sda_high
	ldi		tmp1,250
	rcall	wait
	ldi		tmp1,250
	rcall	wait
	rjmp	error_fpga

; *****************************************************************************

; functions

; *****************************************************************************

; just wait N*1ms, N is given in tmp1

wait:
	ldi		tmp2,20*F_CPU_MHz
wait_1:
	ldi		tmp3,16
wait_2:
	dec		tmp3
	brne	wait_2
	dec		tmp2
	brne	wait_1
	dec		tmp1
	brne	wait
	ret

; *****************************************************************************

; write a byte to I2C
; input  : tmp1   : data
; output : T-flag : ACK (0= NACK (SDA high), 1 = ACK (SDA low))
 
i2c_write:
	ldi		tmp2, 8		; loop conter
i2c_write_loop:

	sbrs	tmp1,7		; copy MSB to SDA
	sda_low
	sbrc	tmp1,7
	sda_high

	delay 14, 8 
	scl_high
	delay 7, 2
	scl_low
	lsl		tmp1		; get next bit
	dec		tmp2
	brne	i2c_write_loop

; check ACK from EEPROM
	sda_high
	delay 14, 5
	scl_high
	clt						; clear T-Flag, = no ACK!
	sbis	cfg_pin,sda		; check SDA is ACK=0 (positive ACK)
	set						; ACK=0, OK
	delay 7, 5
	scl_low
	delay 14, 7		
	ret

; *****************************************************************************

; read a byte from I2C
; input	 : T-flag : ACK, (0= NACK (SDA high), 1 = ACK (SDA low))
; output : tmp1   : data
i2c_read:
	ldi		tmp2, 8			; loop conter
i2c_read_loop:
	scl_high
	clc						; copy SDA to carry
	sbic	cfg_pin, sda	; read SDA
	sec						; set carry
	rol		tmp1			; shift in next bit
	delay 7, 6
	scl_low					; minimum pulse width LOW = 1.3us = 13 clks @10 MHz
	delay 14, 5
	dec		tmp2
	brne	i2c_read_loop

; write ACK to EEPROM

	brtc	i2c_read_no_ack
	sda_low					; write ACK, more to read
	rjmp	i2c_read_send_ack
i2c_read_no_ack:
	sda_high				; no ACK, stop reading
i2c_read_send_ack:
	delay 14, 6
	scl_high
	delay 7, 2
	scl_low
	sda_high				; release SDA
	ret

; *****************************************************************************

; Create a I2C START Condition
  
i2c_start:
	delay 5, 2
	sda_low
	delay 5, 2
	scl_low
	delay 30, 7
	ret

; *****************************************************************************

; Create a I2C STOP Condition

i2c_stop:
	delay 5, 2
	sda_low
	delay 5,2 
	scl_high
	delay 5,2
	sda_high
	delay 30, 7
	ret

; *****************************************************************************

; write a byte to FPGA, with bit reversal
; input  : tmp1   : data

.if bitreverse==1		
.message "Bit reverse ON, slow configuration."	
; configuration uses online bit reverse, slow
; bit reverse, LSB first
fpga_write:
	ldi		tmp2, 8		; loop conter
fpga_write_loop:

	sbrs	tmp1,0		; copy LSB to DIN
	din_low
	sbrc	tmp1,0
	din_high

	cclk_high			; generate cclk pulse
	cclk_low
	lsr		tmp1		; get next bit
	dec		tmp2
	brne	fpga_write_loop
	ret

; *****************************************************************************
.else
.message "Bit reverse OFF, fast configuration."	
; configuration uses offline bit reverse, fast (bitdreher.exe)
; no bit reverse, MSB first
fpga_write:
	ldi		tmp2, 8		; loop conter
fpga_write_loop:

	sbrs	tmp1,7		; copy MSB to DIN
	din_low
	sbrc	tmp1,7
	din_high

	cclk_high			; generate cclk pulse
	cclk_low
	lsl		tmp1		; get next bit
	dec		tmp2
	brne	fpga_write_loop
	ret

; *****************************************************************************
.endif


