--
-- Multi channel PWM generator and 9:1 multiplexer
--  
-- 27 channels with 10 Bit PWM with 32??? nonlinear steps
--
-- PWM runs single shot mode, used for LED MUXING
--

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

Library UNISIM;
use UNISIM.vcomponents.all;


entity pwm_27 is
    Port (  clk       : in std_logic;                       -- system clock , 8MHz
            trigger   : in std_logic;                       -- trigger new MUX cycle in sync mode
            sync_mode : in std_logic;                       -- sync mode
            pwm       : out std_logic_vector(26 downto 0);  -- PWM outputs, high active
            digits    : out std_logic_vector(8 downto 0);   -- Digits for MUXing
            pwm_adr   : out std_logic_vector(7 downto 0);   -- address for PWM data RAM
            dec_data  : in std_logic_vector(12 downto 0)    -- nonlinear decoded PWM value
         );

end pwm_27;

architecture Behavioral of pwm_27 is

type pwm_array_t is array(26 downto 0) of std_logic_vector(12 downto 0);
type state_t is (s_reset, s_reload_start, s_reload_1, s_reload_loop, s_reload_2, s_reload_3, s_run, s_next_digit);

signal pwm_cnt          : std_logic_vector(13 downto 0);        -- PWM counter
signal pwm_val          : pwm_array_t;                          -- PWM counter array
signal state            : state_t;                              -- State for FSM

signal pwm_sel          : std_logic_vector(26 downto 0);        -- one hot select for PWM counter reload
signal dig_sel          : std_logic_vector(8 downto 0);         -- one hot select for digit MUX
signal chan_cnt         : std_logic_vector(7 downto 0);         -- PWM channel counter
signal loop_cnt         : std_logic_vector(4 downto 0);         -- loop counter, for 27 PWM channels
signal run              : std_logic;                            -- PWM cycle running

begin

-- signal assignments

digits <= dig_sel;
pwm_adr <= chan_cnt;

-- PWM counters

process (clk)
begin
  if rising_edge(clk) then
    for i in 0 to 26 loop
      if run='1' and pwm_val(i)/=0 then
        pwm_val(i) <= pwm_val(i) -1;        -- count down with stop at zero
      elsif pwm_sel(i)='1' then
        pwm_val(i) <= dec_data;             -- reload PWM coutner
      end if;
    end loop;
  end if;
end process;

-- PWM output generator

process (clk)
begin
  if rising_edge(clk) then
    for i in 0 to 26 loop
      if pwm_val(i)=0 then
        pwm(i) <='0';         -- clear PWM output
      elsif run='1' then
        pwm(i) <='1';         -- set PWM output
      end if;
    end loop;
  end if;
end process;

-- FSM for PWM Generation and MUXING

process (clk)
begin
  if rising_edge(clk) then
    case state is

      when s_reset =>                               -- start of a complete MUX cycle
        chan_cnt <= "00000000";
        run <= '0';
        if sync_mode='1' or trigger='1' then
          state <= s_reload_start;
          dig_sel  <= "000000001";
        end if;

      when s_reload_start =>                            -- reload PWM counters
        pwm_cnt  <= conv_std_logic_vector(8200, 14);
        chan_cnt <= chan_cnt+1;
        state <= s_reload_1;

      when s_reload_1 =>
        chan_cnt <= chan_cnt+1;
        loop_cnt <= "00000";
        pwm_sel  <= "000000000000000000000000001";
        state <= s_reload_loop;     

      when s_reload_loop =>
        pwm_sel <= pwm_sel(25 downto 0) & '0';          -- shift left, select next PWM counter
        chan_cnt <= chan_cnt+1;
        loop_cnt <= loop_cnt+1;
        if loop_cnt=24 then
          state <= s_reload_2;
        end if;

      when s_reload_2 =>
        pwm_sel <= pwm_sel(25 downto 0) & '0';          -- shift left, select next PWM counter
        loop_cnt <= loop_cnt+1;                         -- last two channels without address increment
        if loop_cnt=26 then                             -- to account for pipelining
          state <= s_reload_3;
        end if;

      when s_reload_3 =>
        state <= s_run;
        run <= '1';

      when s_run =>                                     -- PWM running
        pwm_cnt <= pwm_cnt-1;
        if pwm_cnt = 0 then
          state <= s_next_digit;
          run <= '0';
        end if;

      when s_next_digit =>
        dig_sel <= dig_sel(7 downto 0) & '0';           -- shift left, select next digit
        if dig_sel = "100000000" then
          state <= s_reset;                             -- new complete cycle
        else
          state <= s_reload_start;                      -- next digit
        end if;
        
      when others =>
        state <= s_reset;
    end case;
  end if;
end process;

end Behavioral;

