
-- Top level of DMX-LED-Matrix

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

Library UNISIM;
use UNISIM.vcomponents.all;

entity toplevel_dmx_matrix is
    Port ( clk       : in std_logic;                        -- system clock, 8MHz
           pwm       : out std_logic_vector(26 downto 0);   -- Segments                         
           digits    : out std_logic_vector(8 downto 0);    -- Digits
           rx_data   : in std_logic;                        -- DMX data
           sync_mode : in std_logic;                        -- Sync Mode
           chn_sel   : in std_logic;                        -- DMX channel range select
           test      : out std_logic_vector(4 downto 0));   -- test pads
end toplevel_dmx_matrix;

architecture Behavioral of toplevel_dmx_matrix is

-- component declarations

COMPONENT pwm_27
PORT(
    clk         : IN std_logic;
    trigger     : in std_logic;
    sync_mode   : in std_logic;
    dec_data    : IN std_logic_vector(12 downto 0);          
    pwm         : OUT std_logic_vector(26 downto 0);
    digits      : OUT std_logic_vector(8 downto 0);
    pwm_adr     : OUT std_logic_vector(7 downto 0)
    );
END COMPONENT;

COMPONENT dec_rom
PORT(
    clk     : IN std_logic;
    adr     : IN std_logic_vector(7 downto 0);          
    data    : OUT std_logic_vector(12 downto 0)
    );
END COMPONENT;


COMPONENT data_receiver
PORT(
    clk         : IN std_logic;
    rx_data     : IN std_logic;
    chn_sel     : IN std_logic;
    pwm_data    : OUT std_logic_vector(7 downto 0);
    pwm_adr     : OUT std_logic_vector(7 downto 0);
    page        : OUT std_logic;
    trigger     : OUT std_logic;
    we          : OUT std_logic
    );
END COMPONENT;

-- signal definitions

-- pwm data dual port RAM, 512x8
signal pwm_ram_adr_a    : std_logic_vector(8 downto 0);         -- PWM module
signal pwm_ram_adr_b    : std_logic_vector(8 downto 0);         -- UART
signal pwm_ram_data_a   : std_logic_vector(7 downto 0);         -- PWM module
signal pwm_ram_data_b   : std_logic_vector(7 downto 0);         -- UART
signal pwm_ram_we       : std_logic;                            -- UART

signal dec_rom_data     : std_logic_vector(12 downto 0);        -- ROM for nonlinear PWM transformation
signal page             : std_logic;                            -- page signal for PWM data RAM
signal trigger          : std_logic;                            -- trigger new MUX cycle
signal chn_sel_int      : std_logic;                            -- select DMX channel range

begin

-- signal assignments
-- Test pins
test(0) <= pwm_ram_data_b(0); 
test(1) <= pwm_ram_data_b(1); 
test(2) <= pwm_ram_data_b(6);
test(3) <= pwm_ram_data_b(7);
test(4) <= '0';
 
chn_sel_int <= not chn_sel;

-- component instanciations

-- PWM generator

l_pwm: pwm_27 PORT MAP(
        clk         => clk,
        trigger     => trigger,
        sync_mode   => sync_mode,
        pwm         => pwm,
        digits      => digits,
        pwm_adr     => pwm_ram_adr_a(7 downto 0),
        dec_data    => dec_rom_data
    );

-- PWM data RAM
-- Port A, read only from PWM-module
-- port B, write only, from DMX receiver

   RAMB4_S8_S8_inst : RAMB4_S8_S8
   port map (
      DOA   => pwm_ram_data_a,  
      DOB   => open,  
      ADDRA => pwm_ram_adr_a,
      ADDRB => pwm_ram_adr_b,
      CLKA  => clk,  
      CLKB  => clk,  
      DIA   => "00000000",  
      DIB   => pwm_ram_data_b,  
      ENA   => '1',  
      ENB   => '1',  
      RSTA  => '0',  
      RSTB  => '0',  
      WEA   => '0',  
      WEB   => pwm_ram_we   
   );

-- Decoder ROM

l_dec_rom: dec_rom PORT MAP(
    clk     => clk,
    adr     => pwm_ram_data_a,
    data    => dec_rom_data
    );

-- DMX receiver

l_data_receiver: data_receiver PORT MAP(
    clk      => clk,
    rx_data  => rx_data,
    chn_sel  => chn_sel_int,
    pwm_data => pwm_ram_data_b,
    pwm_adr  => pwm_ram_adr_b(7 downto 0),
    page     => page,
    trigger  => trigger,
    we       => pwm_ram_we
    );

-- PWM data RAM page control

process(sync_mode, page)
begin
  if sync_mode='1' then                 -- MUX free running, data directly written to PWM RAM
    pwm_ram_adr_b(8) <= '0';
    pwm_ram_adr_a(8) <= '0';
  else                                  -- MUX synced to DMX signal
    pwm_ram_adr_b(8) <= page;           -- one page is written by UART
    pwm_ram_adr_a(8) <= not page;       -- while the other is read by PWM module
  end if;
end process;

end Behavioral;
